Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://app/jsmodules/sbLibraryUtils.jsm");
Components.utils.import("resource://app/jsmodules/sbProperties.jsm");
Components.utils.import("resource://app/jsmodules/kPlaylistCommands.jsm");
Components.utils.import("resource://app/jsmodules/ArrayConverter.jsm");
Components.utils.import("resource://app/jsmodules/SBDataRemoteUtils.jsm");

const EXTENSION_UUID = "nowplaying@j.c.m";

const NOWPLAYING_MODE_BASIC = 0;
const NOWPLAYING_MODE_CUSTOM = 1;
const NOWPLAYING_MODE_QUEUE = 2;
const NOWPLAYING_MODE_PARTY = 3;

const NOWPLAYING_LIST_CONTRACTID = "@j.c.m/NowPlaying/List;1";
const NOWPLAYING_LIST_CLASSNAME = "Now Playing List";
const NOWPLAYING_LIST_CID = Components.ID("{7ce97f4b-a1bd-4a90-b53c-49d5b4cd0e1a}");

const SONGBIRD_DATAREMOTE_CONTRACTID = "@songbirdnest.com/Songbird/DataRemote;1";

const SONGBIRD_STRINGBUNDLE = "chrome://songbird/locale/songbird.properties";
const JCM_STRINGBUNDLE = "chrome://jcm-nowplaying/locale/nowplaying.properties";

const LIST_GUID_PREF     = "extensions.nowplaying@j.c.m.listguid";
const QUEUE_GUID_PREF    = "extensions.nowplaying@j.c.m.queueguid";
const HIDE_NODE_PREF     = "extensions.nowplaying@j.c.m.hidenode";
const SHOW_CONTENT_PREF  = "extensions.nowplaying@j.c.m.showcontent";
const FORK_PREF          = "extensions.nowplaying@j.c.m.fork";
const START_MODE_PREF    = "extensions.nowplaying@j.c.m.startmode";

const Cu = Components.utils;
const Cc = Components.classes;
const Ci = Components.interfaces;


var gOS    = null;
var gPrefs = null;
var gPPS   = null;

function NowPlayingList()
{
  gOS    = Cc["@mozilla.org/observer-service;1"]
           .getService(Ci.nsIObserverService);
  gPrefs = Cc["@mozilla.org/preferences-service;1"]
           .getService(Ci.nsIPrefBranch);
  
  if (gOS.addObserver) {
    gOS.addObserver(this, "songbird-library-manager-ready", false);
    gOS.addObserver(this, "songbird-library-manager-before-shutdown", false);
    gOS.addObserver(this, "xul-window-visible", false);
    
    gOS.addObserver(this, "em-action-requested", false);
    gOS.addObserver(this, "quit-application-granted", false);
  }
  
  this.wrappedJSObject = this;
}


// COMPONENT

NowPlayingList.prototype.constructor = NowPlayingList;

NowPlayingList.prototype = {
  classDescription : NOWPLAYING_LIST_CLASSNAME,
  classID          : NOWPLAYING_LIST_CID,
  contractID       : NOWPLAYING_LIST_CONTRACTID,
  
  _xpcom_factory  :
  {
    singleton: null,
    createInstance: function (aOuter, aIID)
    {
      if (aOuter != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      if (this.singleton == null)
        this.singleton = new NowPlayingList();
      return this.singleton.QueryInterface(aIID);
    }
  },
  
  _xpcom_categories : [{
   // Each object in the array specifies the parameters to pass to
   // nsICategoryManager.addCategoryEntry(). 'true' is passed for
   // both aPersist and aReplace params.
   category: "app-startup",
   
   // optional, defaults to the object's classDescription
   //entry: "entry name",
   
   // optional, defaults to the object's contractID (unless
   // 'service' is specified)
   //value: "...",
   
   // optional, defaults to false. When set to true, and only if 'value'
   // is not specified, the concatenation of the string "service," and the
   // object's contractID is passed as aValue parameter of addCategoryEntry.
   service: true
 }],
  
  
  
  MODE_BASIC : NOWPLAYING_MODE_BASIC,
  MODE_CUSTOM : NOWPLAYING_MODE_CUSTOM,
  MODE_QUEUE : NOWPLAYING_MODE_QUEUE,
  MODE_PARTY : NOWPLAYING_MODE_PARTY,
  
  
  _list              : null,
  _mode              : null,
  _playlist          : null,
  
  _queueBookmark     : null,
  _playedQueuedTrack : false,
  _preQueueView      : null,
  _preQueueIndex     : null,
  
  // basic mode attaches a listener to track filter and sort changes
  _isListener        : false,
  
  _timer             : null,
  
  // Data Remotes
  _label             : null,
  _length            : null,
  _canSave           : null,
  _canRandomise      : null,
  _canClear          : null,
  
  // fix for timing conflicts between bindPlaylist and _init
  _initialised     : false,
  _playlistWaiting : null,
  
  // prevent double initialisation
  _initialised2 : false,
  
  // for extension cleanup
  _uninstall : false,
    
  get list() {
    return this._list;
  },
  
  get mode() {
    return this._mode;
  },
  
  get playlist() {
    return this._playlist;
  },
  
  
  _init : function() {
    if (this._initialised) return;
    
    /* Setup list */
    var guid = gPrefs.getCharPref(LIST_GUID_PREF);
    var library = LibraryUtils.mainLibrary;
    try {
      var item = library.getMediaItem(guid);
      var list = item.QueryInterface(Ci.sbIMediaList);
      this._list = list;
    } catch (e) {
      // Doesn't exist
      var list = library.createMediaList("simple");
      list.name = getString("playlistname", "Now playing");
      // save the media list
      gPrefs.setCharPref(LIST_GUID_PREF, list.guid);
      // set up default column specifications
      var columnSpec = SBProperties.trackName + " 170 " +
                       SBProperties.rating + " 80";
      list.setProperty(SBProperties.columnSpec, columnSpec);
      list.setProperty(SBProperties.defaultColumnSpec, columnSpec);
      this._list = list;
    }
    
    this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    
    // attach Data Remotes
    var createDataRemote = new Components.Constructor( SONGBIRD_DATAREMOTE_CONTRACTID, Ci.sbIDataRemote, "init");
    this._label        = createDataRemote("jcm.nowplaying.label", null);
    this._length       = createDataRemote("jcm.nowplaying.length", null);
    this._canSave      = createDataRemote("jcm.nowplaying.cansave", null);
    this._canRandomise = createDataRemote("jcm.nowplaying.canrandomise", null);
    this._canClear     = createDataRemote("jcm.nowplaying.canclear", null);
    
    // attach listeners
    gPPS.addListener(this);
    
    // initialise
    this._mode = gPrefs.getIntPref(START_MODE_PREF);
    
    this._initialised = true;
    if (this._playlistWaiting) this.bindPlaylist(this._playlistWaiting);
  },
  
  _init2 : function() {
    if (this._initialised2) return;
    
    var LSPS  = Cc['@songbirdnest.com/servicepane/library;1']
                .getService(Ci.sbILibraryServicePaneService);
    var SPS   = Cc['@songbirdnest.com/servicepane/service;1']
                .getService(Ci.sbIServicePaneService);
    
    // set up the service pane nodes for the list and queue
    var node = LSPS.getNodeForLibraryResource(this._list);
    node.hidden = gPrefs.getBoolPref("extensions.nowplaying@j.c.m.hidenode");
    
    this._queueBookmark = SPS.addNode('http://nowplayinglist/queue/', SPS.root, false);
    if (this._queueBookmark) {
      this._queueBookmark.url = 'chrome://jcm-nowplaying/content/xul/queueDummyPage.xul';
      this._queueBookmark.name = 'Play Queue';
      this._queueBookmark.image = 'chrome://jcm-nowplaying/content/icons/favicon.png';
      this._queueBookmark.hidden = false;
      this._queueBookmark.editable = false;
    } else {
      this._queueBookmark = SPS.getNode('http://nowplayinglist/queue/');
      this._queueBookmark.name = formatQueueName(0);
    }
    // position node
    var libraryNode = LSPS.getNodeForLibraryResource(LibraryUtils.mainLibrary);
    SPS.root.insertBefore(this._queueBookmark, libraryNode);
    
    this._initialised2 = true;
  },
  
  _deinit : function() {
    if (this._list) {
      this._list.clear();
    }
    this._list = null;
    
    this._queueBookmark = null;
    this._preQueueView = null;
    
    // Cancel the timer just incase it is running.  
    // We don't want it to fire after shutdown.
    this._timer.cancel();
    this._timer = null;
    
    // release Data Remotes
    this._label.unbind();
    this._length.unbind();
    this._canSave.unbind();
    this._canRandomise.unbind();
    this._canClear.unbind();
    
    // release listeners
    gPPS.removeListener(this);
    
    // if haven't already
    if (this._playlist) this.unbindPlaylist();
    
    // save prefs
    gPrefs.setIntPref(START_MODE_PREF, this._mode);
  },
  
  _doUninstall : function() {
    // delete playlist
    var library = LibraryUtils.mainLibrary;
    var item = this._list.QueryInterface(Ci.sbIMediaItem);
    library.remove(item);
    
    // delete queue node
    var SPS   = Cc['@songbirdnest.com/servicepane/service;1']
                .getService(Ci.sbIServicePaneService);
    SPS.removeNode(this._queueBookmark);
    
    // clear preferences
    gPrefs.clearUserPref(LIST_GUID_PREF);
  },
  
  
  
  bindPlaylist : function(aPlaylist) {
    if (!this._initialised) {
      this._playlistWaiting = aPlaylist;
      return;
    }
    
    if (this._playlist) return;
    
    this._playlist = aPlaylist;
    switch (this._mode) {
    case NOWPLAYING_MODE_BASIC :
      this._toBasic();
      break;
    case NOWPLAYING_MODE_CUSTOM :
      this._toCustom();
      break;
    case NOWPLAYING_MODE_QUEUE :
      this._toQueue();
      break;
    case NOWPLAYING_MODE_PARTY :
      // TODO
      break;
    }
    
    this._update();
    
    this._playlistWaiting = null;
  },
  
  unbindPlaylist : function() {
    // release listeners
    if (this._isListener) {
      this._isListener = false;
      gPPS.playingView.removeListener(this);
    }
    
    this._playlist = null;
  },
  
  
  
  setMode : function(newMode) {
    if (this._isListener) {
      this._isListener = false;
      gPPS.playingView.removeListener(this);
    }
    switch (newMode) {
    case NOWPLAYING_MODE_BASIC :
      this._toBasic();
      break;
    case NOWPLAYING_MODE_CUSTOM :
      this._toCustom();
      break;
    case NOWPLAYING_MODE_QUEUE :
      this._toQueue();
      break;
    case NOWPLAYING_MODE_PARTY :
      // TODO
    default :
      this.setMode(NOWPLAYING_MODE_BASIC);
      break;
    }
    this._update();
  },
  
  /*
  Basic mode binds whatever media list view playing to the now playing list.
  If there is nothing playing or the show content pref is off, it will empty
  the custom media list and bind that to the playlist.
  */
  _toBasic : function() {
    this._mode = NOWPLAYING_MODE_BASIC;
    
    if (gPPS.playing &&
        gPPS.playingView.mediaList.equals(this._currentList()) &&
        gPrefs.getBoolPref(SHOW_CONTENT_PREF)) {
      return;
    }
    
    if (gPPS.playing && gPrefs.getBoolPref(SHOW_CONTENT_PREF)) {
      // can't use cloneView() because it keeps filter constraints
      var view = gPPS.playingView.mediaList.createView();
      this._bindView(view);
      
      gPPS.playingView.addListener(this)
      this._isListener = true;
      this._syncSearchAndFilterConstraints(gPPS.playingView);
    } else {
      this._list.clear();
      
      var list = this._currentList();
      if (list && list.equals(this._list)) return;
      
      var view = this._list.createView();
      this._bindView(view);
    }
  },
  
  /*
  Custom mode binds a 'hidden' playlist to the now playing list which can be
  edited 'on the go'.
  */
  _toCustom : function() {
    this._mode = NOWPLAYING_MODE_CUSTOM;
    
    var list = this._currentList();
    if (list && list.equals(this._list)) return;
    
    var view = this._list.createView();
    this._bindView(view);
  },
  
  /*
  Queue mode is the same as custom mode, except that as long as the list is not empty,
  the next song played is the track in the first position, which gets deleted after
  finishing playing. Play then resumes as normal to the original view.
  */
  _toQueue : function() {
    this._mode = NOWPLAYING_MODE_QUEUE;
    
    var list = this._currentList();
    if (list && list.equals(this._list)) return;
    
    var view = this._list.createView();
    this._bindView(view);
  },
  
  _bindView : function(aView) {
    if (!this._playlist) return;
    
    var commandsMgr = Cc["@songbirdnest.com/Songbird/PlaylistCommandsManager;1"]
                        .createInstance(Ci.sbIPlaylistCommandsManager);
    var commands = commandsMgr.request(kPlaylistCommands.MEDIAITEM_DEFAULT);
    this._playlist.bind(aView, commands);
    this._playlist.getListView().selection.selectNone()
  },
  
  _fillList : function() {
    this._list.clear();
    
    if (!gPPS.playing) return;
    if (gPPS.playingView.mediaList.equals(this._list)) return;
    
    var view = gPPS.playingView;
    var items = [];
    for (var i = 0; i < view.length; i++) {
      // javascript is slow ...
      items.push(view.getItemByIndex(i));
    }
    var enum = ArrayConverter.enumerator(items);
    this._list.addSome(enum);
  },
  
  
  
  /* sbIPlaylistPlaybackListener */
  onStop: function() {
    // only do this if playback has really stopped
    this._timer.init(this, 500, Ci.nsITimer.TYPE_ONE_SHOT);
  },
  _stopTimeout: function() {
    if (gPPS.playing) return;
    
    var currentList = this._currentList();
    
    switch (this._mode) {
    case NOWPLAYING_MODE_BASIC :
      this._list.clear();
      if (currentList && currentList.equals(this._list)) return;
      this._bindView(this._list.createView());
      break;
      
    case NOWPLAYING_MODE_CUSTOM :
      return;
      
    case NOWPLAYING_MODE_QUEUE :
      if (this._playedQueuedTrack) {
        this._playedQueuedTrack = false;
        this._list.clear();
        // restore original view
        gPPS.playView(this._preQueueView, this._preQueueIndex);
      }
      break;
      
    }
    
    this._update();
  },
  
  /* sbIPlaylistPlaybackListener */
  onBeforeTrackChange: function(aItem, aView, aIndex) {
    
  },
  
  /* sbIPlaylistPlaybackListener */
  onTrackChange: function(aItem, aView, aIndex) {
    
    switch (this._mode) {
    case NOWPLAYING_MODE_BASIC :
      // will return if there's nothing to do
      // otherwise binds the correct media list view to the playlist
      // if it changed
      this._toBasic();
      break;
    case NOWPLAYING_MODE_CUSTOM :
      break;
    case NOWPLAYING_MODE_QUEUE :
      this._playQueuedTrack(aItem, aView, aIndex);
      break;
    }
    
    this._update();
  },
  
  _playQueuedTrack : function(aItem, aView, aIndex) {
    // BUG: need to check why the track changed
    // wait for play history service to be implemented
    // basically if user doubleclicked on track, do nothing here
    
    if (this._list.isEmpty) return;
    
    if (aView.mediaList.equals(this._list)) {
      
      if (this._playedQueuedTrack) {
        // remove previously played track
        this._list.removeByIndex(0);
        
        // play the track at the top, even if shuffle is on, this is a queue
        this._playedQueuedTrack = false;
        gPPS.playView(aView, 0);
      } else {
        this._playedQueuedTrack = true;
      }
      
    } else if (gPPS.repeat != gPPS.REPEAT_MODE_ONE) {
      
      // save current view and index
      this._preQueueView = aView;
      this._preQueueIndex = aIndex;
      
      // first track from queue
      this._playedQueuedTrack = false;
      gPPS.playView(this._list.createView(), 0);
      
    }
  },
  
  
  
  /* sbIMediaListViewListener */
  onFilterChanged: function(aChangedView) {
    this._update();
    this._syncSearchAndFilterConstraints(aChangedView);
  },
  
  /* sbIMediaListViewListener */
  onSearchChanged: function(aChangedView) {
    this._update();
    this._syncSearchAndFilterConstraints(aChangedView);
  },
  
  /* sbIMediaListViewListener */
  onSortChanged: function(aChangedView) { },
  
  
  
  _update : function() {
    
    var queueLength = 0;
    
    switch (this._mode) {
    case NOWPLAYING_MODE_BASIC :
      this._canSave.boolValue      = false;
      this._canRandomise.boolValue = false;
      this._canClear.boolValue     = false;
      
      if (gPPS.playing) {
        var view = gPPS.playingView;
        this._label.stringValue = view.mediaList.name;
        this._length.stringValue = formatLength(view.length);
        
        // librarys don't support ordering, and clearing should be disabled
        // to prevent accidental deletion
        // but not if fork is on
        if (gPrefs.getBoolPref(SHOW_CONTENT_PREF)) {
          if ( !(view.mediaList instanceof Ci.sbILibrary) || gPrefs.getBoolPref(FORK_PREF) ) {
            this._canRandomise.boolValue = true;
            this._canClear.boolValue     = true;
          }
        }
        
        // most cases just returns immediately
        // important when user is turning off show content pref
        this._toBasic();
        
      } else {
        this._label.stringValue = "";
        this._length.stringValue = "";
      }
      break;
      
    case NOWPLAYING_MODE_CUSTOM :
      this._label.stringValue = getString("customplaylist", "Custom playlist");
      this._length.stringValue = formatLength(this._list.length);
      
      this._canSave.boolValue      = true;
      this._canRandomise.boolValue = true;
      this._canClear.boolValue     = true;
      break;
      
    case NOWPLAYING_MODE_QUEUE :
      this._label.stringValue = getString("queuename", "Play Queue");
      this._length.stringValue = formatLength(this._list.length);
      queueLength = this._list.length;
      
      this._canSave.boolValue      = true;
      this._canRandomise.boolValue = false;
      this._canClear.boolValue     = true;
      break;
      
    }
    
    if (this._queueBookmark) {
      this._queueBookmark.name = formatQueueName(queueLength);
    }
  },
  
  _syncSearchAndFilterConstraints : function(aSyncView) {
    var list  = this._currentList();
    if (!list) return;
    if (list.equals(this._list)) return;
    
    try {
      var search = aSyncView.searchConstraint;
      var filter = aSyncView.filterConstraint;
      
      var view = this._playlist.getListView();
      view.searchConstraint = search;
      view.filterConstraint = filter;
    } catch (e) {
      dump("Now Playing: syncSearchAndFilterConstraints - " + e);
    }
  },
  
  
  
  /* Actions */
  
  save : function() {
    // Don't use LibraryUtils as we don't know which library the playlist is
    // from
    var library = this._currentList().library;
    var newList = library.copyMediaList("simple", this._currentList());
    newList.name = getString("customplaylist", "Custom playlist");
    
    var LSPS  = Cc['@songbirdnest.com/servicepane/library;1']
                .getService(Ci.sbILibraryServicePaneService);
    var node = LSPS.getNodeForLibraryResource(newList);
    node.name = getString("customplaylist", "Custom playlist");
    
    // We don't ask for a new name anymore
  },
  
  randomise : function() {
    if ((this._mode == NOWPLAYING_MODE_BASIC) && 
        gPrefs.getBoolPref(FORK_PREF)) {
      this._fillList();
      this._toCustom();
      this._update();
    }
    
    var mediaList = this._currentList();
    var n = mediaList.length;
    var orderable = mediaList.QueryInterface(Ci.sbIOrderableMediaList);
    while (--n >= 0) {
      var k = Math.round(n*Math.random());
      orderable.moveLast(k);
    }
    
    // wierd stuff gets selected after randomising
    this._playlist.getListView().selection.selectNone();
  },
  
  clear : function() {
    
    switch (this.mode) {
    case this.MODE_BASIC :
      if (gPrefs.getBoolPref("extensions.nowplaying@j.c.m.fork")) {
        this._list.clear();
        this._toCustom();
      } else {
        if (!(this._currentList() instanceof Ci.sbILibrary)) {
          this._currentList().clear();
        }
      }
      
      // Clearing a library when fork is off is disabled
      // The warning dialog is too complicated to put here and won't
      // work inside XPCOM
      // not having one would be too easy for the user to
      // delete the entire library
      
      break;
    case this.MODE_CUSTOM :
    case this.MODE_QUEUE :
      this._list.clear();
    }
    
    this._update();
  },
  
  enqueue : function(items) {
    this._list.addSome(items);
    this._toQueue();
    this._update();
  },
  
  
  
  _currentList : function() {
    try {
      return this._playlist.getListView().mediaList;
    } catch (e) {
      return null;
    }
  },
  
  
  
  // watch for XRE startup and shutdown messages 
  observe: function(subject, topic, data) {
    switch (topic) {
    case "timer-callback" :
      this._stopTimeout();
      this._timer.cancel();
      break;
    case "songbird-library-manager-ready" :
      gOS.removeObserver(this, "songbird-library-manager-ready")  
      /* Other services */
      gPPS = Cc["@songbirdnest.com/Songbird/PlaylistPlayback;1"]
             .getService(Ci.sbIPlaylistPlayback);
      // Preferences are initialized, ready to start the service
      this._init();
      break;
    case "xul-window-visible" :
      gOS.removeObserver(this, "xul-window-visible");
      this._init2();
      break;
    // Must shut down the service before the library goes away
    case "songbird-library-manager-before-shutdown" :
      gOS.removeObserver(this, "songbird-library-manager-before-shutdown");
      
      this._deinit();
      
      // Release Services to avoid memory leaks
      gOS    = null;
      gPrefs = null;
      gPPS   = null;
      break;
    case "em-action-requested" :
      subject.QueryInterface(Components.interfaces.nsIUpdateItem);
      if (subject.id == EXTENSION_UUID) {
        switch (data) {
        case "item-uninstalled" :
        case "item-disabled" :
          this._uninstall = true;
          break;
        case "item-cancel-action" :
          this._uninstall = false;
          break;
        }
      }
      break;
    case "quit-application-granted" :
      gOS.removeObserver(this,"em-action-requested");
      gOS.removeObserver(this,"quit-application-granted");
      if (this._uninstall) {
        this._doUninstall();
      }
      break;
    }
  },
  
  QueryInterface:
    XPCOMUtils.generateQI([Ci.nsIObserver, Ci.sbIPlaylistPlaybackListener, Ci.sbIMediaListViewListener]) 
};



// XPCOM Registration
function NSGetModule(compMgr, fileSpec) {
  return XPCOMUtils.generateModule([NowPlayingList]);
}


// helper functions

function getString(aName, aDefaultValue) {
  var stringBundleService = Cc["@mozilla.org/intl/stringbundle;1"]
                            .getService(Ci.nsIStringBundleService);
  
  var sbBundle = stringBundleService.createBundle(SONGBIRD_STRINGBUNDLE);
  try {
    return sbBundle.GetStringFromName(aName);
  } catch (e) {} // do nothing, try extension string bundle
  
  var jcmBundle = stringBundleService.createBundle(JCM_STRINGBUNDLE);
  try {
    return jcmBundle.GetStringFromName(aName);
  } catch(e) {
    return aDefaultValue;
  }
}

function formatLength(aLength) {
  switch (aLength) {
    case 0 :
      return getString("library.noitem", "no items"); break;
    case 1 :
      return "1 " + getString("library.oneitem", "item"); break;
    case "" :
      return ""; break;
    default :
      return aLength + " " + getString("library.manyitems", "items"); break;
  }
}

function formatQueueName(aLength) {
  var queueName = getString("queuename", "Play Queue");
  switch (aLength) {
  case 0 :
    return queueName;
  default :
    return queueName + " (" + aLength + ")";
  }
}
